#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Script de métriques utiles pour les approches online

Created on Wed Nov  6 11:26:53 2019

@author: Cyrile.Delestre
"""
import warnings
from typing import (Union, Iterable, Dict, Any, Optional, NewType, List,
                    Callable)
from functools import wraps

import numpy as np
from pandas import DataFrame, Series
from sklearn.metrics import (mean_absolute_error, median_absolute_error,
                             mean_squared_error, accuracy_score, roc_auc_score,
                             precision_score, precision_recall_fscore_support)

from ._base import BaseEnvironnement

PyTorchEstimators = NewType('PyTorchEstimators', BaseEnvironnement)

def mean_absolute_error_online(estimator: PyTorchEstimators,
                               X: Iterable,
                               y_true: Iterable,
                               kargs_fit: Dict[str, Any],
                               gap: int=0,
                               idx_skip: Optional[Union[np.ndarray,
                                                        List[int]]]=None,
                               **kargs):
    r"""
    Fonction mean_absolute_error appliquée au contexte online. Voici la 
    documentation Scikit-Learn :
        `mean_absolute_error`_
    
    .. _mean_absolute_error: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_absolute_error.html
    
    Parameters
    ----------
    estimator : PyTorchEstimators
        objet estimator de type regressor
    X : Iterable
        observations
    y_true : Iterable
        targets associées aux observations X
    kargs_fit : Dict[str, Any]
        dictionnaire des arguments nécéssaires au fit
    gap : int
        Permet de définir un gap entre l'apprentissage du modèle et la 
        prédiction. Par défaut 0, aucun gap. Dans la pratique le gap est à 
        minima égal à l'horizon de prédiction.
    idx_skip : Optional[Union[np.ndarray, List[int]]]
        indices des observation à exclure. Permet de limiter l'effet des black 
        swan pour cetain type de série.
    **kargs :
        paramètres associés à la fonction Scikit-Learn (voir la doc 
        Scikit-Learn).
    """
    check_estimator(
        estimator=estimator,
        estimator_type='regressor',
        fun_score='mean_absolute_error_online'
    )
    y_pred = estimator.fit_online(
        X,
        predict = True,
        gap = gap,
        **kargs_fit
    )
    y_true, y_pred = skip_obs(y_true[gap:], y_pred, idx_skip, gap)
    return apply_metric_error(
        mean_absolute_error,
        y_true,
        y_pred,
        **kargs
    )

def median_absolute_error_online(estimator: PyTorchEstimators,
                                 X: Iterable,
                                 y_true: Iterable,
                                 kargs_fit: Dict[str, Any],
                                 gap: int=0,
                                 idx_skip: Optional[Union[np.ndarray,
                                                          List[int]]]=None,
                                 **kargs):
    r"""
    Fonction median_absolute_error appliquée au contexte online. Voici la 
    documentation Scikit-Learn :
        `median_absolute_error`_
    
    .. _median_absolute_error: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.median_absolute_error.html?highlight=median%20absolute
    
    Parameters
    ----------
    estimator : PyTorchEstimators
        objet estimator de type regressor
    X : Iterable
        observations
    y_true : Iterable
        targets associées aux observations X
    kargs_fit : Dict[str, Any]
        dictionnaire des arguments nécéssaires au fit
    gap : int
        Permet de définir un gap entre l'apprentissage du modèle et la 
        prédiction. Par défaut 0, aucun gap. Dans la pratique le gap est à 
        minima égal à l'horizon de prédiction.
    idx_skip : Optional[Union[np.ndarray, List[int]]]
        indices des observation à exclure. Permet de limiter l'effet des black 
        swan pour cetain type de série.
    **kargs :
        paramètres associés à la fonction Scikit-Learn (voir la doc 
        Scikit-Learn).
    """
    check_estimator(
        estimator=estimator,
        estimator_type='regressor',
        fun_score='median_absolute_error_online'
    )
    y_pred = estimator.fit_online(
        X,
        predict=True,
        gap=gap,
        **kargs_fit
    )
    y_true, y_pred = skip_obs(y_true[gap:], y_pred, idx_skip, gap)
    return apply_metric_error(
        median_absolute_error,
        y_true,
        y_pred,
        **kargs
    )

def mean_squared_error_online(estimator: PyTorchEstimators,
                              X: Iterable,
                              y_true: Iterable,
                              kargs_fit: Dict[str, Any],
                              gap: int=0,
                              idx_skip: Optional[Union[np.ndarray,
                                                       List[int]]]=None,
                              **kargs):
    r"""
    Fonction mean_squared_error appliquée au contexte online. Voici la 
    documentation Scikit-Learn :
        `mean_squared_error`_
    
    .. _mean_squared_error: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_squared_error.html?highlight=mean_squared_error
    
    Parameters
    ----------
    estimator : PyTorchEstimators
        objet estimator de type regressor
    X : Iterable
        observations
    y_true : Iterable
        targets associées aux observations X
    kargs_fit : Dict[str, Any]
        dictionnaire des arguments nécéssaires au fit
    gap : int
        Permet de définir un gap entre l'apprentissage du modèle et la 
        prédiction. Par défaut 0, aucun gap. Dans la pratique le gap est à 
        minima égal à l'horizon de prédiction.
    idx_skip : Optional[Union[np.ndarray, List[int]]]
        indices des observation à exclure. Permet de limiter l'effet des black 
        swan pour cetain type de série.
    **kargs :
        paramètres associés à la fonction Scikit-Learn (voir la doc 
        Scikit-Learn).
    """
    check_estimator(
        estimator=estimator,
        estimator_type='regressor',
        fun_score='mean_squared_error_online'
    )
    y_pred = estimator.fit_online(
        X,
        predict = True,
        gap = gap,
        **kargs_fit
    )
    y_true, y_pred = skip_obs(y_true[gap:], y_pred, idx_skip, gap)
    return apply_metric_error(
        mean_squared_error,
        y_true,
        y_pred,
        **kargs
    )

def accuracy_score_online(estimator: PyTorchEstimators,
                          X: Iterable,
                          y_true: Iterable,
                          kargs_fit: Dict[str, Any],
                          gap: int=0,
                          idx_skip: Optional[Union[np.ndarray,
                                                   List[int]]]=None,
                          **kargs):
    r"""
    Fonction accuracy_score appliquée au contexte online. Voici la 
    documentation Scikit-Learn :
        `accuracy_score`_
    
    .. _accuracy_score: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.accuracy_score.html?highlight=accuracy_score
    
    Parameters
    ----------
    estimator : PyTorchEstimators
        objet estimator de type regressor
    X : Iterable
        observations
    y_true : Iterable
        targets associées aux observations X
    kargs_fit : Dict[str, Any]
        dictionnaire des arguments nécéssaires au fit
    gap : int
        Permet de définir un gap entre l'apprentissage du modèle et la 
        prédiction. Par défaut 0, aucun gap. Dans la pratique le gap est à 
        minima égal à l'horizon de prédiction.
    idx_skip : Optional[Union[np.ndarray, List[int]]]
        indices des observation à exclure. Permet de limiter l'effet des black 
        swan pour cetain type de série.
    **kargs :
        paramètres associés à la fonction Scikit-Learn (voir la doc 
        Scikit-Learn).
    """
    check_estimator(
        estimator=estimator,
        estimator_type='classifier',
        fun_score='accuracy_score_online'
    )
    y_pred = estimator.fit_online(
        X,
        predict=True,
        gap=gap,
        **kargs_fit
    )
    y_true, y_pred = skip_obs(y_true[gap:], y_pred, idx_skip, gap)
    return accuracy_score(y_true, y_pred, **kargs)

def roc_auc_score_online(estimator: PyTorchEstimators,
                         X: Iterable,
                         y_true: Iterable,
                         kargs_fit: Dict[str, Any],
                         keys: int=1,
                         gap: int=0,
                         idx_skip: Optional[Union[np.ndarray,
                                                  List[int]]]=None,
                         **kargs):
    r"""
    Fonction roc_auc_score appliquée au contexte online. Voici la 
    documentation Scikit-Learn :
        `roc_auc_score`_
    
    .. _roc_auc_score: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.roc_auc_score.html?highlight=roc_auc_score

    Parameters
    ----------
    estimator : PyTorchEstimators
        objet estimator de type regressor
    X : Iterable
        observations
    y_true : Iterable
        targets associées aux observations X
    kargs_fit : Dict[str, Any]
        dictionnaire des arguments nécéssaires au fit
    keys : int
        colonne de sortie de prédiction correspondant à la classe testé
    gap : int
        Permet de définir un gap entre l'apprentissage du modèle et la 
        prédiction. Par défaut 0, aucun gap. Dans la pratique le gap est à 
        minima égal à l'horizon de prédiction.
    idx_skip : Optional[Union[np.ndarray, List[int]]]
        indices des observation à exclure. Permet de limiter l'effet des black 
        swan pour cetain type de série.
    **kargs :
        paramètres associés à la fonction Scikit-Learn (voir la doc 
        Scikit-Learn).
    """
    check_estimator(
        estimator=estimator,
        estimator_type='classifier',
        fun_score='roc_auc_score_online'
    )
    y_pred = estimator.fit_online(
        X,
        predict = True,
        needs_proba = True,
        gap = gap,
        **kargs_fit
    )
    if y_pred.shape[1] > 2:
        warnings.warn(
            "Attention, l'estimateur est multi-classes (pas bi-classes)."
        )
    y_true, y_pred = skip_obs(y_true[gap:], y_pred, idx_skip, gap)
    y_score = y_pred[:, keys]
    return roc_auc_score(y_true, y_score, **kargs)

def precision_score_online(estimator: PyTorchEstimators,
                           X: Iterable,
                           y_true: Iterable,
                           kargs_fit: Dict[str, Any],
                           gap: int=0,
                           idx_skip: Optional[Union[np.ndarray,
                                                    List[int]]]=None,
                           **kargs):
    r"""
    Fonction precision_score appliquée au contexte online. Voici la 
    documentation Scikit-Learn :
        `precision_score`_
    
    .. _precision_score: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.precision_score.html?highlight=precision_score
    
    Parameters
    ----------
    estimator : PyTorchEstimators
        objet estimator de type regressor
    X : Iterable
        observations
    y_true : Iterable
        targets associées aux observations X
    kargs_fit : Dict[str, Any]
        dictionnaire des arguments nécéssaires au fit
    gap : int
        Permet de définir un gap entre l'apprentissage du modèle et la 
        prédiction. Par défaut 0, aucun gap. Dans la pratique le gap est à 
        minima égal à l'horizon de prédiction.
    idx_skip : Optional[Union[np.ndarray, List[int]]]
        indices des observation à exclure. Permet de limiter l'effet des black 
        swan pour cetain type de série.
    **kargs :
        paramètres associés à la fonction Scikit-Learn (voir la doc 
        Scikit-Learn).
    """
    check_estimator(
        estimator=estimator,
        estimator_type='classifier',
        fun_score='precision_score_online'
    )
    y_pred = estimator.fit_online(
        X,
        predict = True,
        gap = gap,
        **kargs_fit
    )
    y_true, y_pred = skip_obs(y_true[gap:], y_pred, idx_skip, gap)
    return precision_score(y_true, y_pred, **kargs)

def precision_recall_fscore_support_online(estimator: PyTorchEstimators,
                                           X: Iterable,
                                           y_true: Iterable,
                                           kargs_fit: Dict[str, Any],
                                           gap: int=0,
                                           idx_skip: Optional[
                                               Union[np.ndarray,
                                                     List[int]]
                                               ]=None,
                                           **kargs):
    r"""
    Fonction precision_recall_fscore_support appliquée au contexte online. 
    Voici la documentation Scikit-Learn :
        `precision_recall_fscore_support`_
    
    .. _precision_recall_fscore_support: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.precision_recall_fscore_support.html?highlight=precision_recall_fscore_support
    
    Parameters
    ----------
    estimator : PyTorchEstimators
        objet estimator de type regressor
    X : Iterable
        observations
    y_true : Iterable
        targets associées aux observations X
    kargs_fit : Dict[str, Any]
        dictionnaire des arguments nécéssaires au fit
    gap : int
        Permet de définir un gap entre l'apprentissage du modèle et la 
        prédiction. Par défaut 0, aucun gap. Dans la pratique le gap est à 
        minima égal à l'horizon de prédiction.
    idx_skip : Optional[Union[np.ndarray, List[int]]]
        indices des observation à exclure. Permet de limiter l'effet des black 
        swan pour cetain type de série.
    **kargs :
        paramètres associés à la fonction Scikit-Learn (voir la doc 
        Scikit-Learn).
    """
    check_estimator(
        estimator=estimator,
        estimator_type='classifier',
        fun_score='precision_recall_fscore_support_online'
    )
    y_pred = estimator.fit_online(
        X,
        predict = True,
        gap = gap,
        **kargs_fit
    )
    y_true, y_pred = skip_obs(y_true[gap:], y_pred, idx_skip, gap)
    return precision_recall_fscore_support(
        y_true, 
        y_pred,
        **kargs
    )

def check_estimator(estimator: PyTorchEstimators,
                    estimator_type: str,
                    fun_score: Optional[str]=None):
    r"""
    Fonction de test du type d'estimateur
    
    Parameters
    ----------
    estimator : PyTorchEstimators
        objet estimateur
    estimator_type : str
        type de l'estimateur souhaité
    fun_score : Optional[str]
        nom de la fonction testant l'estimateur. Par défaut None.
    """
    typ = ['regressor', 'classifier']
    if not estimator_type in typ:
        raise ValueError(
            f"Le type estimator_type: {estimator_type} doit être "
            f"{', '.join(typ)}."
        )
    if estimator._estimator_type != estimator_type:
        if fun_score is None:
            raise ValueError(
                f"L'estimateur entré n'est pas de type {estimator_type} mais "
                f"de type {estimator._estimator_type}. Il faut un "
                f"{estimator_type} pour utiliser cette métrique."
            )
        else:
            raise ValueError(
                f"L'estimateur entré de {fun_score} n'est pas de type "
                f"{estimator_type} mais de type {estimator._estimator_type}. "
                f"Il faut un {estimator_type} pour utiliser cette métrique."
            )

def reverse_score(loss_fun: Callable, *args, **kargs):
    r"""
    Décorateur d'inversement d'une fonction coût. Les fonctions en entrée de 
    l'argument scoring de la classe RandomizedSearchOnline sont considérées 
    comme étant des fonctions score par Scikit-Learn. Si la fonction de mesure 
    de performance est une fonction coût, il faut inverser les résultats afin 
    que le score le plus élevé corresponde au résultat de la fonction coût le 
    plus bas. Dans Scikit-Learn ce cas est traité en mutipliant le résultat 
    par -1 ce qui permet de préserver le sens physique de la fonction fonction 
    coût (il suffira de multiplier le résultat par -1). Cette stratégie est 
    préservée à l'aide de cette fonction.
    
    Parameters
    ----------
    loss_fun : Callable
        fonction coût à inverser
    *args :
        se référer à la fonction loss_fun utilisée ou custom
    **kargs :
        se référer à la fonction loss_fun utilisée ou custom
    """
    @wraps(loss_fun)
    def wrapper(*args, **kargs):
        res = loss_fun(*args, **kargs)
        return -1*res
    return wrapper

def apply_metric_error(fun: Callable,
                       y_true: Iterable,
                       y_pred: Iterable,
                       **kargs):
    r"""
    Application de la métrique fun. Si dans y_pred il y a un NaN ou un inf 
    alors retourne comme score Inf.
    
    Parameters
    ----------
    fun : Callable
        fonciton coût ou de score
    y_true : Iterable
        réalité terrain
    y_pred : Iterable
        prédiction
    **kargs:
        arguments de la fonction fun
    
    Returns
    -------
    score : float
        score issu de fun ou inf.
    """
    if all(~np.isnan(y_pred)) and all(~np.isinf(y_pred)):
        return fun(y_true, y_pred, **kargs)
    else:
        return np.inf

def skip_obs(y_true: Iterable,
             y_pred: Iterable,
             idx_skip: Optional[Iterable],
             gap: int):
    r"""
    Fonction permettant d'extraire certaines observations de l'évaluation.
    
    Parameters
    ----------
    y_true : Iterable
        array-like de la réalitée terrain
    y_pred : Iterable
        estimation de l'agorithme de ML
    idx_skip : Optional[Iterable]
        array-like indices des observation à exclure
    gap : int
        gap appliquer à l'évaluation
    """
    if idx_skip is None:
        idx_skip = []
    if not isinstance(idx_skip, np.ndarray):
        idx_skip = np.array(idx_skip)
    idx_skip = idx_skip[idx_skip > gap]-gap
    idx = np.arange(0, len(y_true))
    idx = idx[~np.isin(idx, idx_skip)]
    y_pred = y_pred[idx, :]
    if isinstance(y_true, list):
        y_true = [y_true[ii] for ii in idx]
    elif isinstance(y_true, np.ndarray):
        y_true = y_true[idx]
    elif isinstance(y_true, (DataFrame, Series,)):
        y_true = y_true.iloc[idx]
    else:
        raise AttributeError(
            "Le type de y_true n'est pas reconnu. Un array-like est soit "
            "de type list, ndarray, DataFrame ou Series, et non de type "
            f"{type(y_true)}."
        )
    return y_true, y_pred



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ._chunk import Chunker, chunker
from ._date import date_inf, date_sup, convert_date_dataframe
from ._timeout import timeout
from ._check import check_dataframe, shape_list
from ._set_deep_params import set_params_deep
from ._statistique import weighted_avg_and_std
from ._transpose import (list_of_dict_2_dict_of_list,
                         dict_of_list_2_list_of_dict)
from ._rolling import RollingWindow

__all__ = (
    'Chunker',
    'chunker',
    'date_inf',
    'date_sup',
    'convert_date_dataframe',
    'timeout',
    'check_dataframe',
    'shape_list',
    'set_params_deep',
    'weighted_avg_and_std',
    'list_of_dict_2_dict_of_list',
    'dict_of_list_2_list_of_dict',
    'RollingWindow'
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ._transformers import Transformer, make_transformer
from ._prepro import (OneHotEncoderPandas, KBinsDiscretizerPandas,
                      RobustScalerPandas, QuantileTransformerPandas)

__all__ = (
    'Transformer',
    'make_transformer',
    'OneHotEncoderPandas',
    'KBinsDiscretizerPandas',
    'RobustScalerPandas',
    'QuantileTransformerPandas'
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Classe de rolling sur une fenêtre
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Created on Sat Mar 27 10:17:24 2021

@author: Cyrile Delestre
"""

from typing import List, Union, Any
from dataclasses import dataclass

import numpy as np
import pandas as pd


@dataclass
class RollingWindow:
    r"""
    Classe permettant de faire un rolling sur un itérateur qui peut être 
    une DataFrame Pandas, un array Numpy ou un liste.

    Parameters
    ----------
    data : Union[List[Any], np.ndarray, pd.DataFrame]
        Itérable où sera appliquer le rolling.
    window : int
        Taille de la fenêtre du rolling. Elle doit être au minimum de taille 
        2 et maximum la taille de data.
    poisition : str (='right')
        Position du centrage de la fenêtre, elle peut être "right", "center" 
        ou "left".

    Notes
    -----
    La classe convertie les types de la DataFrame d'origine aux type étandu 
    de Pandas 1.0. La classe n'est pas compatible avec une version de 
    Pandas inférieur. Une copy de la sous DataFrame est faite en sortie 
    d'ésoladirisant ainsi le retour du rolling avec la DataFrame d'origine.

    Examples
    --------
    >>> import pandas as pd
    >>> from dstk.utils import RollingWindow
    >>> 
    >>> df = pd.DataFrame({'a': ['a', 'b', 'c', 'd'], 'b': [1,2,3,4]})
    >>> roll = RollingWindow(test,3, position='right')
    >>> roll[0]
          a     b
    0  <NA>  <NA>
    1  <NA>  <NA>
    2     a     1
    >>> roll[1]
          a     b
    0  <NA>  <NA>
    1     a     1
    2     b     2
    >>> len(roll)
    4
    >>> roll[4]
    IndexError: Index out of band, l'index doit être compris entre 0 et 3.
    """
    data: Union[List[Any], np.ndarray, pd.DataFrame]
    window: int
    position: str = 'right'

    def __post_init__(self):
        if not isinstance(self.data, (list, np.ndarray, pd.DataFrame)):
            raise AttributeError(
                "data doit être un array-like (list, NumPy array ou "
                f"DataFrame) et non de type {type(self.data)}."
            )
        if self.position not in ['left', 'right', 'center']:
            raise AttributeError(
                "La position de la fenêtre doit être 'left', 'center' ou "
                f"'right' et non '{self.position}'."
            )
        if self.window > len(self.data):
            raise AttributeError(
                "La taille de la fenêtre ne peut pas être supérieur à la "
                "taille de la donnée."
            )
        if isinstance(self.data, pd.DataFrame):
            self.pandas = True
            self.data = self.data.convert_dtypes()
        else: self.pandas = False

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        if idx > self.__len__()-1:
            raise IndexError(
                "Index out of band, l'index doit être compris entre 0 et "
                f"{self.__len__()-1}."
            )
        if self.position == 'right':
            if idx < self.window-1:
                return self._out_of_lower_band(idx, 1)
            else:
                return (
                    self.data[idx-self.window+1:idx+1].copy()
                    if not self.pandas else
                    (
                        self.data
                        .iloc[idx-self.window+1:idx+1]
                        .copy()
                    )
                )
        elif self.position == 'left':
            if idx > len(self.data)-self.window:
                return self._out_of_upper_band(idx, 0)
            else:
                return (
                    self.data[idx:idx+self.window].copy()
                    if not self.pandas else
                    (
                        self.data
                        .iloc[idx:idx+self.window]
                        .copy()
                    )
                )
        else:
            middle_win = (self.window-1)//2
            right_res = self.window-middle_win
            if idx < middle_win:
                return self._out_of_lower_band(idx, self.window//2+1)
            elif idx > len(self.data)-right_res:
                return self._out_of_upper_band(idx, (self.window-1)//2)
            else:
                return (
                    self.data[idx-middle_win:idx+right_res].copy()
                    if not self.pandas else
                    (
                        self.data
                        .iloc[idx-middle_win:idx+right_res]
                        .copy()
                    )
                )

    def _out_of_lower_band(self, idx: int, offset: int):
        if self.pandas:
            roll = pd.DataFrame(
                columns=self.data.columns,
                index=(
                    [pd.NA]*(self.window-idx-offset)
                    +list(self.data.iloc[0:idx+offset].index)
                )
            )
            roll.iloc[self.window-idx-offset:self.window] = (
                self.data.iloc[0:idx+offset]
            )
            roll = roll.astype(self.data.dtypes.to_dict())
            return roll
        elif isinstance(self.data, np.ndarray):
            roll = np.empty((self.window, self.data.shape[1]))
            roll.fill(np.nan)
            roll[self.window-idx-offset:self.window] = (
                self.data[0:idx+offset]
            )
            return roll
        else:
            return self.data[0:idx+offset].copy()

    def _out_of_upper_band(self, idx: int, offset: int):
        if self.pandas:
            roll = pd.DataFrame(
                columns=self.data.columns,
                index=(
                    list(self.data.iloc[idx-offset:].index)
                    +[pd.NA]*(self.window-len(self.data)+idx-offset)
                )
            )
            roll.iloc[0:len(self.data)-idx+offset] = (
                self.data.iloc[idx-offset:]
            )
            roll = roll.astype(self.data.dtypes.to_dict())
            return roll
        elif isinstance(self.data, np.ndarray):
            roll = np.empty((self.window, self.data.shape[1]))
            roll.fill(np.nan)
            roll[0:len(self.data)-idx+offset] = self.data[idx-offset:]
            return roll
        else:
            return self.data[idx-offset:].copy()



import os
try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup

    def find_packages(where='.'):
        # os.walk -> list[(dirname, list[subdirs], list[files])]
        return [folder.replace("/", ".").lstrip(".")
                for (folder, _, fils) in os.walk(where)
                if "__init__.py" in fils]

import builtins

builtins.__SETUP__ = True

import dstk

with open("README.md") as readme_file:
    readme = readme_file.read()

requirements = [
    "joblib",
    "numpy",
    "scikit-learn",
    "pandas",
    "intel-openmp!=2019.5",
    "tqdm"
]

setup(
    name="dstk_x250",
    version=dstk.__version__,
    platforms=["any"],
    packages=["dstk"]+[f"dstk.{ii}" for ii in find_packages("dstk")],
    url="https://gitlab.com/Kirire/x250",
    python_requires=">=3.5",
    include_package_data = True,
    license="Apache Software License 2.0",
    author="Cyrile Delestre",
    author_email="cyrile.ufr.orsay@gmail.com",
    description="Package d'utilitaires pour les projets de data science.",
    install_requires=requirements,
    setup_requires=["pytest-runner"],
    test_suite="tests",
    tests_require=["pytest"],
    long_description=open(os.path.join(os.path.dirname(__file__), 'README.md')).read(),
    long_description_content_type='text/markdown',
    zip_safe=False
)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Transposition de dictionnaires
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Created on Wed Nov 25 11:37:41 2020

@author: Cyrile Delestre
"""

from typing import (List, Tuple, Union, Dict, Optional, Callable, Iterable,
                    NewType)

import numpy as np

TypeListOfDict = NewType(
    'TypeListOfDict',
    Union[
        List[Dict[str, Union[str, float, int]]],
        Tuple[Dict[str, Union[str, float, int]]]
    ]
)

def list_of_dict_2_dict_of_list(data: TypeListOfDict,
                                keys_trans: Optional[List[str]]=None,
                                type_trans: Callable=np.asarray):
    r"""
    Fonciton transposant un tuple ou une liste de dictionnaire en un 
    dictionnaire d'itérateur de type type_trans.
    
    Parameters
    ----------
    data : TypeListOfDict
        liste ou tuple de dictionnaire
    keys_trans : Optional[List[str]]
        clés des arguments du dico souhaité à être transposé, par défaut None 
        (toutes les clefs)
    type_trans : Callable
        type de la transposition, par défaut numpy array
    
    Examples
    --------
    >>> from dstk.utils import list_of_dict_2_dict_of_list as l2d
    
    >>> list_of_dict
        [{'a': 1, 'b': 4}, {'a': 2, 'b': 5}, {'a': 3, 'b': 6}]
    
    >>> l2d(list_of_dict, type_trans=list)
        {'a': [1, 2, 3], 'b': [4, 5, 6]}
    
    See also
    --------
    dict_of_list_2_list_of_dict
    """
    if not isinstance(data, (tuple, list)):
        raise ValueError(
            "data doit être du type tuple ou liste, et non de type "
            f"{type(data)}."
        )
    if keys_trans is None:
        keys_trans = data[0].keys()
    transpose = {
        key: type_trans(
            [dd[key] for dd in data]
        ) for key in keys_trans
    }
    return transpose

def dict_of_list_2_list_of_dict(data: Dict[str, Iterable],
                                keys_trans: Optional[List[str]]=None):
    r"""
    Fonction transposant un dictionnaire d'itérable en une liste de 
    dictionnaire.
    
    Parameters
    ----------
    data : Dict[str, Iterable]
        dictionnaire d'itérables que l'on souhaite transposer
    keys_trans : Optional[List[str]]
        clés des arguments du dico souhaité à être transposé, par défaut None 
        (toutes les clefs)
    
    Examples
    --------
    >>> from dstk.utils import dict_of_list_2_list_of_dict as d2l
    
    >>> dict_of_array
        {'a': array([1, 2, 3]), 'b': array([4, 5, 6])}
    
    >>> d2l(dict_of_array)
        [{'a': 1, 'b': 4}, {'a': 2, 'b': 5}, {'a': 3, 'b': 6}]
    
    See also
    --------
    dict_of_list_2_list_of_dict
    """
    if not isinstance(data, dict):
        raise ValueError(
            f"data doit être de type dict et non de type {type(data)}."
        )
    if keys_trans is None:
        keys_trans = data.keys()
    data = {kk: data[kk] for kk in keys_trans}
    kk, vv = zip(*data.items())
    transpose = [
        dict(zip(kk, ii)) for ii in zip(*vv)
    ]
    return transpose



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Classe pour wrapper facilement un modèle de machine learning personnel avec 
Scikit-Learn.

Created on Mon Nov 23 13:31:25 2020

@author: Cyrile Delestre
"""

from collections.abc import Iterable
from typing import Optional

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from joblib import Parallel, delayed
from tqdm import tqdm

class Base(BaseEstimator):
    r"""
    Classe générique de BaseEstimator Scikit-Learn qui implémente une
    exécution compatible avec les Pipeline où les éléments doivent être
    traités un à un si ils sont dans une liste ou un itérable. Elle est
    compatible pour les Classifier et les Regressor. Elle nécessite
    l'implémentation de udf_fit(X, y, **kargs) et
    udf_predict(self, X, **kargs).
    
    Parameters
    ----------
    n_jobs : int
        nombre de processeurs en parallèle
    verbose : bool
        Si True bare de progression de process.
    
    Notes
    -----
    Cette classe n'est pas à utiliser directement (sauf si nécessaire). Il
    est préférable de se reporter sur la classe Regressor s'il s'agit d'un
    regresseur ou Classifier s'il s'agit d'un classifieur.
    
    See also
    --------
    Regressor, Classifier
    """
    def __init__(self, n_jobs: int, verbose: bool):
        self.n_jobs = n_jobs
        self.verbose = n_jobs

    def _fit_iter(self, X: Iterable, y: Iterable, **kargs):
        r"""
        Itération de l'apprentissage.
        """
        fit = delayed(self.udf_fit)
        if y is None:
            iterator = tqdm(
                X, 
                desc = f"{self.__class__.__name__}",
                mininterval = 0.5
            ) if self.verbose else X

            with Parallel(n_jobs=self.n_jobs) as par:
                par(fit(x, **kargs) for x in iterator)
        elif isinstance(y, Iterable):
            iterator = tqdm(
                zip(X, y),
                desc = f"{self.__class__.__name__}",
                mininterval = 0.5
            ) if self.verbose else zip(X, y)

            with Parallel(n_jobs=self.n_jobs) as par:
                par(fit(*x, **kargs) for x in iterator)
        else:
            raise AttributeError(
                "y doit être soit None et dans ce cas X doit contenir la "
                f"target, soit être un Iterable, ici y est de type {type(y)}."
            )

    def udf_fit(self, X: Iterable, y: Iterable, **kargs):
        r"""
        Fit unitaire.
        """
        raise NotImplementedError(
            "Il faut implémenter la méthode udf_fit(self, X, y, **kargs) à "
            f"la classe {self.udf_fit.__name__} qui hérite de la classe "
            "Transformer."
        )

    def fit(self, X: Iterable, y: Optional[Iterable]=None, **kargs):
        r"""
        Fonction fit pour être ISO avec Scikit-Learn
        
        Parameters
        ----------
        X : Iterable
            pour un traitement itératif un Iterable
        y : Optional[Iterable]
            target de l'entrainement.
        **kargs :
            arguments propres au transformer
        """
        if isinstance(X, (pd.DataFrame, np.ndarray,)):
            self.udf_fit(X, y, **kargs)
        elif isinstance(X, Iterable):
            self._fit_iter(X, y, **kargs)
        else:
            raise AttributeError(
                "L'argument X de fit doit être un Iterable."
            )
        return self

    def _predict_iter(self, X: Iterable, **kargs):
        r"""
        Itération de la prédiction.
        """
        iterator = tqdm(
            X,
            desc = f"{self.__class__.__name__}",
            mininterval = 0.5
        ) if self.verbose else X

        pred = delayed(self.udf_predict)
        with Parallel(n_jobs=self.n_jobs) as par:
            return par(pred(x, **kargs) for x in iterator)

    def udf_predict(self, X: Iterable, **kargs):
        r"""
        Prédiction unitaire.
        """
        raise NotImplementedError(
            "Il faut implémenter la méthode udf_predict(self, X, **kargs) à "
            f"la classe {self.udf_predict.__name__} qui hérite de la classe "
            "Transformer."
        )

    def predict(self, X: Iterable, **kargs):
        r"""
        Fonction predict pour être ISO avec Scikit-Learn
        
        Parameters
        ----------
        X : Iterable
            pour un traitement itératif un Iterable
        **kargs :
            arguments propres au transformer
        """
        if isinstance(X, (pd.DataFrame, np.ndarray,)):
            return self.udf_predict(X, **kargs)
        elif isinstance(X, Iterable):
            return self._predict_iter(X, **kargs)
        else:
            raise AttributeError(
                "L'argument X de predict doit être un Iterable."
            )

class Regressor(Base, RegressorMixin):
    r"""
    Classe générique Regressor héritant de la classe Base.
    
    Classe générique de BaseEstimator Scikit-Learn qui implémente une
    exécution compatible avec les Pipeline où les éléments doivent être
    traités un à un si ils sont dans une liste ou un itérable. Elle nécessite
    l'implémentation de udf_fit(X, y, **kargs) et
    udf_predict(self, X, **kargs).
    
    Parameters
    ----------
    n_jobs : int
        nombre de processeurs en parallèle
    verbose : bool
        Si True bare de progression de process.
    
    Notes
    -----
    A utiliser si l'algorithme de régression n'est pas standard à
    Scikit-Learn. Par exemple Gensim, Tensorflow, PyTorch, etc.
    
    A noter que pour PyTorch il est conseillé d'utiliser dstk.pytorch 
    wrapper d'un modèle PyTorch à Scikit-Learn.
    
    Examples
    --------
    Considérons que je possède un modèle SuperAlgoRegressor qui n'est pas
    prototypé Scikit-Learn.
    
    >>> from dataclasses import dataclass
    >>> from sklearn.preprocessing import RobustScaler
    >>> from sklearn.pipeline import Pipeline
    >>> ...
    >>> from mon_projet import SuperAlgoRegressor
    >>> from dstk.ml import Regressor
    >>> 
    >>> @dataclass
    >>> class SuperAlgo2Sklear(Regressor):
    >>>     "Classe SuperAlgoRegressor vers Scikit-Learn."
    >>>     model: SuperAlgoRegressor
    >>>     n_jobs: int=1
    >>>     verbose: bool=False
    >>> 
    >>>     def udf_fit(self, X, y, lr=1e-3):
    >>>         self.model.train(X, y, lr)
    >>> 
    >>>     def udf_predict(self, X, weights = None):
    >>>         return self.model.estimation(X, weights)
    
    >>> scal = RobustScaler()
    >>> mon_model = SuperAlgoRegressor(**kargs_de_mon_SuperAlgoRegressor)
    >>> reg = SuperAlgo2Sklear(mon_model, verbose=True)
    >>> pipe = Pipeline([('scale', scal), ('regressor', reg)])
    >>> pipe.fit(X, y, lr=1e-5)
    
    See also
    --------
    Classifier, dstk.pytorch
    """
    def __init__(self, n_jobs: int, verbose: bool):
        self.n_jobs = n_jobs
        self.verbose = verbose

class Classifier(Base, ClassifierMixin):
    r"""
    Classe générique Classifier héritant de la classe Base.
    
    Classe générique de BaseEstimator Scikit-Learn qui implémente une
    exécution compatible avec les Pipeline où les éléments doivent être
    traités un à un si ils sont dans une liste ou un itérable. Elle nécessite
    l'implémentation de udf_fit(X, y, **kargs) et udf_predict(X, **kargs)
    et udf_predict_proba(X, **kargs).
    
    Parameters
    ----------
    n_jobs : int
        nombre de processeurs en parallèle
    verbose : bool
        Si True bare de progression de process.
    
    Notes
    -----
    A utiliser si l'algorithme de classification n'est pas standard à
    Scikit-Learn. Par exemple Gensim, Tensorflow, PyTorch, etc.
    
    A noter que pour PyTorch il est conseillé d'utiliser dstk.pytorch wrapper 
    d'un modèle PyTorch à Scikit-Learn.
    
    Examples
    --------
    Considérons que je possède un modèle SuperAlgoClassifier qui n'est pas
    prototypé Scikit-Learn.
    
    >>> from dataclasses import dataclass
    >>> from sklearn.preprocessing import RobustScaler
    >>> from sklearn.pipeline import Pipeline
    >>> ...
    >>> from mon_projet import SuperAlgoClassifier
    >>> from dstk.ml import Classifier
    >>> 
    >>> @dataclass
    >>> class SuperAlgo2Sklear(Classifier):
    >>>     "Classe SuperAlgoClassifier vers Scikit-Learn."
    >>>     model: SuperAlgoClassifier
    >>>     n_jobs: int=1
    >>>     verbose: bool=False
    >>> 
    >>>     def udf_fit(self, X, y, lr=1e-3):
    >>>         self.model.train(X, y, lr)
    >>> 
    >>>     def udf_predict(self, X, weights = None):
    >>>         return self.model.estimation(X, weights)
    >>> 
    >>>     def udf_predict_proba(self, X, weights = None)
    >>>         return self.model.estimation(X, weights, proba=True)
    
    >>> scal = RobustScaler()
    >>> mon_model = SuperAlgoClassifier(**kargs_de_mon_SuperAlgoClassifier)
    >>> clf = SuperAlgo2Sklear(mon_model, verbose=True)
    >>> pipe = Pipeline([('scale', scal), ('classifier', clf)])
    >>> pipe.fit(X, y, lr=1e-5)
    
    See also
    --------
    Regressor, dstk.pytorch
    """
    def __init__(self, n_jobs: int, verbose: bool):
        self.n_jobs = n_jobs
        self.verbose = verbose

    def _predict_proba_iter(self, X, **kargs):
        r"""
        Itération de la prédiction.
        """
        iterator = tqdm(
            X,
            desc = f"{self.__class__.__name__}",
            mininterval = 0.5
        ) if self.verbose else X

        pred = delayed(self.udf_predict_proba)
        with Parallel(n_jobs=self.n_jobs) as par:
            return par(pred(x, **kargs) for x in iterator)

    def udf_predict_proba(self, X: Iterable, **kargs):
        r"""
        Prédiction unitaire.
        """
        raise NotImplementedError(
            "Il faut implémenter la méthode udf_predict_proba(self, X, **kargs) "
            f"à la classe {self.udf_predict_proba.__name__} qui hérite de la "
            "classe Transformer."
        )

    def predict_proba(self, X: Iterable, **kargs):
        r"""\
        Fonction predict_proba pour être ISO avec Scikit-Learn

        Parameters
        ----------
        X : Iterable
            DataFrame, pour un traitement itératif un Iterable
        **kargs :
            arguments propres au transformer
        """
        if isinstance(X, (pd.DataFrame, np.ndarray,)):
            return self.udf_predict_proba(X, **kargs)
        elif isinstance(X, Iterable):
            return self._predict_proba_iter(X, **kargs)
        else:
            raise AttributeError(
                "L'argument X de predict_proba doit être un Iterable."
            )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ._handlers import (SQLiteHandler, SQLAlchemyHandler, ArtimonAPIHandler, 
                        ArtimonLocalHandler)
from ._aggregator import Aggregator
from ._metrics import (timer, timer_class, sniffer, sniffer_class)

__all__ = (
    'SQLiteHandler',
    'SQLAlchemyHandler',
    'ArtimonAPIHandler',
    'ArtimonLocalHandler',
    'Aggregator',
    'ClientSSH',
    'timer',
    'timer_class',
    'sniffer',
    'sniffer_class',
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Classe et fonction des chunkers
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Created on Mon Nov 23 09:38:09 2020

@author: Cyrile Delestre
"""

from dataclasses import dataclass
from typing import Any, Union, List
from math import ceil, floor

import pandas as pd
import numpy as np


@dataclass
class Chunker:
    r"""
    Classe permettant de tronçonner une liste data d'objets en N paquets de 
    taille size.
    
    Parameters
    ----------
    data : Union[List[Any], np.ndarray, pd.DataFrame]
        array-like ou liste d'éléments
    size : int
        taille d'un tronçon
    drop_remainder : bool
        élimine le dernier tronçon si ce dernier n'est pas de taille size (si 
        len(data) n'est pas divisible par size).
    
    Notes
    -----
    Convient si la donnée ne prend pas trop de place sur la RAM sinon utiliser 
    la fonction "chuncker".
    
    Examples
    --------
    >>> from dstk.utils import Chunker
    >>> 
    >>> chunk = Chunker(list(range(1000)), 10)
    >>> len(chunk)
    100
    >>> chunk[5]
    [50, 51, 52, 53, 54, 55, 56, 57, 58, 59]
    >>> chunk[100]
    IndexError: Index out of band, l'index doit être compris entre 0 et 99.
    
    See also
    --------
    chuncker
    """
    data: Union[List[Any], np.ndarray, pd.DataFrame]
    size: int
    drop_remainder: bool=False
    def __post_init__(self):
        if not isinstance(self.data, (list, np.ndarray, pd.DataFrame)):
            raise AttributeError(
                "data doit être un array-like (list, NumPy array ou "
                f"DataFrame) et non de type {type(self.data)}."
            )
        self.pandas = False
        if isinstance(self.data, pd.DataFrame):
            self.pandas = True
    
    def __len__(self):
        if self.drop_remainder:
            return floor(len(self.data)/self.size)
        else:
            return ceil(len(self.data)/self.size)

    def __getitem__(self, idx):
        if idx > self.__len__()-1:
            raise IndexError(
                "Index out of band, l'index doit être compris entre 0 et "
                f"{self.__len__()-1}."
            )
        return (
            self.data[idx*self.size: idx*self.size+self.size]
            if not self.pandas else
            self.data.iloc[idx*self.size: idx*self.size+self.size]
        )
    

def chunker(data: Union[List[Any], np.ndarray, pd.DataFrame],
            size: int,
            drop_remainder: bool=False):
    r"""
    Chunker pour de grosses volumétries. Génère un générateur consommant  
    aucun espace RAM supplémentaire (par rapport à la data source).
    
    Parameters
    ----------
    data : Union[List[Any], np.ndarray, pd.DataFrame]
        array-like ou liste d'éléments
    size : int
        taille d'un tronçon
    drop_remainder : bool
        élimine le dernier tronçon si ce dernier n'est pas de taille size (si 
        len(data) n'est pas divisible par size).
    
    Notes
    -----
    Si la volumétrie le permet, utiliser plutôt la classe Chunker qui facilite 
    la manipulation du chunker.
    
    Examples
    --------
    >>> from dstk.utils import chunker
    >>> 
    >>> chunk = chunker(range(1000), 10)
    >>> next(chunk)
    range(0, 10)
    >>> next(chunck)
    range(10, 20)
    >>> ...
    >>> next(chunck)
    StopIteration
    
    See also
    --------
    Chuncker
    """
    if isinstance(data, (list, np.ndarray)):
        if drop_remainder:
            for ii in range(0, len(data)-(len(data)%size), size):
                yield data[ii:ii+size]
        else:
            for ii in range(0, len(data), size):
                yield data[ii:ii+size]
    elif isinstance(data, pd.DataFrame):
        if drop_remainder:
            for ii in range(0, len(data)-(len(data)%size), size):
                yield data.iloc[ii:ii+size]
        else:
            for ii in range(0, len(data), size):
                yield data.iloc[ii:ii+size]
    else:
        raise AttributeError(
            "data doit être un array-like (list, NumPy array ou DataFrame) "
            f"et non de type {type(data)}."
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Fonctions décoratrice permettant de réaliser une mesure.

Created on Sun Nov 22 12:29:49 2020

@author: Cyrile Delestre
"""

import logging
from functools import wraps
from typing import Optional, Union, List
from datetime import datetime as dt

from . import Aggregator
from .push import _push_logger

def timer(logger: Union[logging.Logger, List, None]=None,
          delta_time: bool=False,
          name: Optional[str]=None):
    r"""
    Calcul de temps d'exécution d'une fonction. Empackage dans un 
    dictionnaire le nom de la fonction (name) et le temps d'éxécution de la 
    fonction en micro-secondes (dt).
    
    Parameters
    ----------
    logger : Union[logging.Logger, List, None]
        Logger, None ou list, si None print dans le terminal, si list 
        append dans la liste (idéalement elle doit être initialement vide) 
        sinon  utilise le logger.
    delta_time : bool
        Si True renvoit au format lisible par un humain
        (HH:MM:SS.microsecondes), si False renvoit le temps d'éxécution en 
        microsecondes.
    name : Optional[str]
        nom de l'identifiant de l'appellant, si None utilise le nom de la 
        fonction.
    
    Examples
    --------
    >>> import logging
    >>> from time import sleep
    >>> from dstk.metrics import timer
    >>> 
    >>> # Création du logger
    >>> logging.basicConfig(
    >>>     level = logging.INFO,
    >>>     style = '{',
    >>>     format = '{name} {message}',
    >>>     filename = 'metrics.log',
    >>>     filemode = 'w'
    >>> )
    >>> logger = logging.getLogger('metrics.python.time')
    
    Envoi vers un logger.
    
    >>> @timer(logger=logger, delta_time=True)
    >>> def function_test(a=1, b=1, sleep_time=0.5):
    >>>     r"Docstring de la fonction fonction_test"
    >>>     sleep(sleep_time)
    >>>     return a+b
    >>> 
    >>> # Envoit dans le fichier metrics.log le nom du logger + le message de 
    >>> # time, par exemple :
    >>> # metrics.python.time {"name": "function_test", "dt": "0:00:00.500830"}
    >>> function_test()
    2
    
    Envoi vers une liste (à noter que dans ce cas où delta_time est égal à 
    True alors dt est un objet datetime.timedelta et non un string) :
    
    >>> log_time = []
    >>> @timer(logger=log_time, delta_time=True)
    >>> def function_test(a=1, b=1, sleep_time=0.5):
    >>>     r"Docstring de la fonction fonction_test"
    >>>     sleep(sleep_time)
    >>>     return a+b
    >>> 
    >>> function_test()
    2
    >>> log_time
    [{"name": "function_test", "dt": 0:00:00.500830}]
    
    Affichage dans la console Python :
    
    >>> @timer(delta_time=True)
    >>> def function_test(a=1, b=1, sleep_time=0.5):
    >>>     r"Docstring de la fonction fonction_test"
    >>>     sleep(sleep_time)
    >>>     return a+b
    >>> 
    >>> function_test()
    {"name": "function_test", "dt": 0:00:00.500830}
    2
    
    See also
    --------
    timer_class
    """
    def decorator(fun):
        @wraps(fun)
        def wrap(*args, **kargs):
            t1 = dt.now()
            out_fun = fun(*args, **kargs)
            d_out = dt.now()-t1
            message = dict(
                name = fun.__name__ if name is None else name,
                dt = d_out if delta_time else d_out.microseconds
            )
            if not isinstance(logger, list) and delta_time:
                message['dt'] = str(message['dt'])
            _push_logger(logger, message)
            return out_fun
        return wrap
    return decorator

def timer_class(logger: Union[logging.Logger, List, None]=None,
                delta_time: bool=False,
                name: Optional[str]=None):
    r"""
    Décorateur d'une classe. Permet de calculer le temps d'exécution de  
    chaque méthode de la classe. Empackage dans un dictionnaire le 
    nom de la méthode (name) et le temps d'éxécution de la méthode en 
    micro-seconde (dt).
    
    Parameters
    ----------
    logger :  Union[logging.Logger, List, None]
        Logger, None ou list, si None print dans le terminal, si list 
        append dans la liste (idéalement elle doit être initialement vide ) 
        sinon  utilise le logger.
    delta_time : bool
        Si True renvoit au format lisible par un humain
        (HH:MM:SS.microsecondes), si False renvoit le temps d'éxécution en 
        microsecondes.
    name : Optional[str]
        nom de l'identifiant de l'appellant, si None utilise le nom de la 
        méthode.
    
    Examples
    --------
    >>> import logging
    >>> from time import sleep
    >>> from dstk.metrics import timer_class
    >>> 
    >>> # Création du logger au format CSV avec séparateur ';'
    >>> logging.basicConfig(
    >>>     level = logging.INFO,
    >>>     style = '{',
    >>>     format = '{asctime};{name};{message}',
    >>>     filename = 'metrics.log',
    >>>     filemode = 'w'
    >>> )
    >>> logger = logging.getLogger('metrics.python.time')
    >>> 
    >>> @timer_class(logger=logger, delta_time=True)
    >>> class ClassTest:
    >>>     r"Docstring de la classe ClassTest"
    >>>     def __init__(self, a=1, b=2):
    >>>         self.a = a
    >>>         self.b = b
    >>> 
    >>>     def method_test(self):
    >>>         r"Docstring de la méthode method_test"
    >>>         sleep(1)
    >>>         return self.a + self.b
    >>> 
    >>>     def method_test_with_input(self, a=1):
    >>>         r"Docstring de la méthode method_test_with_input"
    >>>         sleep(0.5)
    >>>         return self.a + self.b + a
    >>> 
    >>>     def method_test_without_output(self, a=1):
    >>>         r"Docstring de la méthode method_test_without_output"
    >>>         sleep(0.1)
    >>>         self.c = a
    
    >>> # L'instentiation d'une classe n'envoit rien dans le logger.
    >>> test = ClassTest()
    >>> # Envoit dans le fichier metrics.log le timestamp le nom du logger + 
    >>> # le message de time, par exemple :
    >>> # 2020-04-11T09:58:13.990425;metrics.python.time;{"name": 
    >>> # "method_test_with_input", "dt": "0:00:00.500652"}
    >>> test.method_test_with_input()
    4
    
    Pour plus d'exemples voir le décorateur "timer".
    
    See also
    --------
    timer
    """
    def decorator(Cls):
        class WrapClass:
            def __init__(self, *args, **kargs):
                self.new_instance = Cls(*args, **kargs)
                self.__doc__ = self.new_instance.__doc__

            def __getattribute__(self, key):
                try:
                    attribute = super(WrapClass, self).__getattribute__(key)
                except AttributeError:
                    pass
                else:
                    return attribute
                attribute = self.new_instance.__getattribute__(key)
                if callable(attribute):
                    return timer(logger, delta_time, name)(attribute)
                else:
                    return attribute
    
        return WrapClass
    return decorator

def sniffer(logger: Union[logging.Logger, List, None]=None,
            agg: Aggregator=Aggregator(),
            name: Optional[str]=None):
    r"""
    Récolte la sortie d'une fonction permettant de surveiller le comportement 
    de la fonction. Package l'information dans un dictionnaire. Ce 
    dictionnaire contient par défaut au moins le type (type) + des infos qui 
    dépendent du type.
    
    Parameters
    ----------
    logger : Union[logging.Logger, List, None]
        Logger, None ou list, si None print dans le terminal, si list 
        append dans la liste (idéalement elle doit être initialement vide) 
        sinon  utilise le logger.
    agg : Aggregator
        Classe :class:`~dstk.metrics._aggregator.Aggregator` qui package 
        dans un dictionnaire les informations issues de la sortie.
    name : Optional[str]
        nom de l'identifiant de l'appellant, si None utilise le nom de la
        fonction.
    
    Examples
    --------
    Exemple d'envoi vers un fichier log classic :
    
    >>> import logging
    >>> from dstk.metrics import Aggregator, sniffer
    >>> 
    >>> # Création du logger
    >>> logging.basicConfig(
    >>>     level = logging.INFO,
    >>>     style = '{',
    >>>     format = '{created} {name} {message}',
    >>>     filename = 'metrics.log',
    >>>     filemode = 'w'
    >>> )
    >>> logger = logging.getLogger('metrics.python.sniffer')
    >>> 
    >>> @sniffer(logger=logger, agg=Aggregator())
    >>> def function_test(a=1, b=1):
    >>>     r"Docstring de la fonction fonction_test"
    >>>     return a+b
    >>> 
    >>> # Envoit dans le fichier metrics.log le timestamp + le nom du logger +
    >>> # le message de sniffer, par exemple :
    >>> # 1587116257.2187557 metrics.python.sniffer {"name": "function_test",
    >>> # "type": "scalar", "out": 2}
    >>> function_test()
    2
    
    Voici un deuxième exemple où on envoit la donnée dans une base SQLite3 :
    
    >>> import numpy as np
    >>> from dstk.metrics import SQLiteHandler
    >>> from dstk.data import PandasSQLite
    >>> 
    >>> # Création du logger
    >>> logging.basicConfig(
    >>>     level = logging.INFO,
    >>>     handlers = [
    >>>         SQLiteHandler(
    >>>             path_sql = 'metrics.log',
    >>>             format = ['asctime', 'name', 'message', 'threadName']
    >>>         )
    >>>     ]
    >>> )
    >>> logger = logging.getLogger('metrics.python.sniffer')
    >>> 
    >>> @sniffer(logger=logger, agg=Aggregator())
    >>> def function_test(a=500, b=10):
    >>>     r"Docstring de la fonction fonction_test"
    >>>     return np.random.rand(a, b)
    >>> 
    >>> function_test()
    array([[0.28586659, 0.68635657, 0.0214463 , ..., 0.72955093, 0.46374203,
        0.91401919],
       [0.86642687, 0.25809078, 0.05919618, ..., 0.66675736, 0.36540648,
        0.87976802],
       ...,
       [0.27390028, 0.36790765, 0.04724844, ..., 0.9266507 , 0.97582395,
        0.18682153]])
    >>> 
    >>> function_test()
    array([[0.36979266, 0.15393057, 0.4848909 , ..., 0.81696902, 0.70991454,
        0.84205372],
       [0.81409839, 0.99492684, 0.62253569, ..., 0.84153022, 0.87136116,
        0.74612971],
       ...,
       [0.17831675, 0.56979541, 0.37167904, ..., 0.01579573, 0.43899887,
        0.55057923]])
    >>> 
    >>> df = PandasSQLite("metrics.log")
    >>> with df:
    >>>     logs = df("SELECT * FROM logs;")
    >>> 
    >>> logs[['asctime', 'name', 'threadName']]
                              asctime                    name  threadName
        0  2020-04-17T12:12:48.759743  metrics.python.sniffer  MainThread
        1  2020-04-17T12:12:48.763768  metrics.python.sniffer  MainThread
    >>> logs.message
        0    {"name": "function_test", "type": "array", "mean": 0.49331038639861874,...
        1    {"name": "function_test", "type": "array", "mean": 0.4902900945580839, ...
        Name: message, dtype: object
    
    See also
    --------
    sniffer_class
    """
    if not isinstance(agg, Aggregator):
        raise AttributeError(
            "L'argument 'agg' doit être un objet de type Aggregator et non "
            f"de type {type(agg)}."
        )
    def decorator(fun):
        @wraps(fun)
        def wrap(*args, **kargs):
            out_fun = fun(*args, **kargs)
            message = agg.make_message(out_fun)
            message.update(
                name = fun.__name__ if name is None else name
            )
            _push_logger(logger, message)
            return out_fun
        return wrap
    return decorator

def sniffer_class(logger: Union[logging.Logger, List, None]=None,
                  agg: Aggregator=Aggregator,
                  name: Optional[str]=None):
    r"""
    Décorateur d'une classe. Récolte la sortie d'une fonction permettant de 
    surveiller le comportement  de la fonction. Package l'information dans un 
    dictionnaire. Ce dictionnaire contient par défaut au moins le type (type) 
    + des infos qui dépendent du type.
    
    Parameters
    ----------
    logger : Union[logging.Logger, List, None]
        logger, None ou list, si None print dans le terminal, si list 
        append dans la liste (idéalement elle doit être initialement vide) 
        sinon  utilise le logger.
    agg : Aggregator
        Classe :class:`~dstk.metrics._aggregator.Aggregator` qui package dans 
        un dictionnaire les informations issues de la sortie.
    name : Optional[str]
        nom de l'identifiant de l'appellant, si None utilise le nom de la 
        fonction.
    
    Examples
    --------
    >>> import logging
    >>> from dstk.metrics import Aggregator, sniffer_class
    >>> 
    >>> # Création du logger au format CSV avec séparateur ';'
    >>> logging.basicConfig(
    >>>     level = logging.INFO,
    >>>     style = '{',
    >>>     format = '{asctime};{name};{message}',
    >>>     filename = 'metrics.log',
    >>>     filemode = 'w'
    >>> )
    >>> logger = logging.getLogger('metrics.python.sniffer')
    >>> 
    >>> @sniffer_class(logger=logger, agg=Aggregator())
    >>> class ClassTest:
    >>>     r"Docstring de la classe ClassTest"
    >>>     def __init__(self, a=1, b=2):
    >>>         self.a = a
    >>>         self.b = b
    >>> 
    >>>     def method_test(self):
    >>>         r"Docstring de la méthode method_test"
    >>>         return self.a + self.b
    >>> 
    >>>     def method_test_with_input(self, a=1):
    >>>         r"Docstring de la méthode method_test_with_input"
    >>>         return self.a + self.b + a
    >>> 
    >>>     def method_test_without_output(self, a=1):
    >>>         r"Docstring de la méthode method_test_without_output"
    >>>         self.c = a
    
    >>> # L'instanciation d'une classe n'envoit rien dans le logger.
    >>> test = ClassTest()
    >>> # Envoit dans le fichier metrics.log le timestamp le nom du logger + 
    >>> # le message de time, par exemple :
    >>> # 2020-04-11T09:58:13.990425;metrics.python.sniffer;{"name": 
    >>> # "method_test_with_input", "type": "scalar", "out": 4}
    >>> test.method_test_with_input()
    4
    
    Pour plus d'exemples voir le décorateur "sniffer".
    
    See also
    --------
    sniffer
    """
    def decorator(Cls):
        class WrapClass:
            def __init__(self, *args, **kargs):
                self.new_instance = Cls(*args, **kargs)
                self.__doc__ = self.new_instance.__doc__

            def __getattribute__(self, key):
                try:
                    attribute = super(WrapClass, self).__getattribute__(key)
                except AttributeError:
                    pass
                else:
                    return attribute
                attribute = self.new_instance.__getattribute__(key)
                if callable(attribute):
                    return sniffer(logger, agg, name)(attribute)
                else:
                    return attribute

        return WrapClass
    return decorator



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ._pandas_sql import PandasSQL, PandasSQLite, PandasSQLAlchemy
from ._sql import csv2sql, index_sql

__all__ = (
    'PandasSQL',
    'PandasSQLite',
    'PandasSQLAlchemy',
    'csv2sql',
    'index_sql'
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Module Stochastic Weight Averaging (SWA)

Created on Tue Jan 19 06:46:56 2021

@author: Cyrile Delestre
"""

import os
from copy import deepcopy
from dataclasses import dataclass
from typing import List, Dict, Union, Optional, NewType, Any, Callable
from collections import OrderedDict
from itertools import chain
from warnings import warn

import numpy as np
from torch import Tensor, no_grad
from sklearn.model_selection import ShuffleSplit, StratifiedShuffleSplit
from torch.utils.data import Dataset, DataLoader, Subset
from torch.optim.swa_utils import AveragedModel, SWALR, update_bn

from dstk.pytorch import (CallbackInterface, CallbackHandler,
                          ProgressBarCallback, FitState, FitControl,
                          check_tensor)
from dstk.utils.meta_interface import SuperclassDocstring
from dstk.utils.errors import DivergenceError

BATCH_NORM_LAYERS = [
    'BatchNorm1d',
    'BatchNorm2d',
    'BatchNorm3d',
    'GroupNorm',
    'SyncBatchNorm',
    'InstanceNorm1d',
    'InstanceNorm2d',
    'InstanceNorm3d',
    'LayerNorm'
]

# Permet d'éviter une initialisation circulaire via l'import de
# BaseEnvironnement
PyTorchEstimators = NewType('PyTorchEstimators', Any)

@dataclass
class SaveStateSWACallback(CallbackInterface, metaclass=SuperclassDocstring):
    r"""
    Classe de gestion des poids de la méthode SWA permettant de préserver le 
    meilleur modèle durant le processus SWA.

    Parameters
    ----------
    old_res_loss : float
        Valeur loss atteinte pendant la phase d'apprentissage classique.
    path_save : Optional[str] (=None)
        Chemin (absolu ou relatif) où le modèle par early stopping sera 
        préservé. Si None, les paramètres seront sauvegardés en RAM. 
        Cependant il est déconseillé de le faire car les réseaux profonds 
        peuvent être très gourmands en espace RAM. De même, en cas de 
        plantage, un état du réseau pouvant être satisfaisant est préservé 
        sur le disque dur.
    early_stopping_rounds : Optional[int] (=None)
        Nombre d'itérations (epoch) sans amélioration. Si None pas de 
        early stopping, le modèle sauvegardé sera donc le meilleur modèle 
        durant l'ensemble des époques.
    minimize : bool (=True)
        Indique si l'objectif est de minimiser la fonciton coût (True) ou 
        de la maximiser (False).
    normalise : bool (=False)
        Indique si une couche de normalisation est présente dans le modèle. 
        Dans ce cas l'évaluation durant l'apprentissage n'est pas possible.
    verbose : bool (=False)
        Indique les informations de gain de performance.

    Notes
    -----
    Cette classe n'a pas pour but d'être utilisée directement, mais est 
    instanciée dans la classe 
    :mod:`~dstk.pytorch._swa.StochasticWeightAveraging`.
    """
    old_res_loss: float
    path_save: Optional[str] = None
    early_stopping_rounds: Optional[int] = None
    minimize: bool = True
    normalise: bool = False
    verbose: bool = False

    def __post_init__(self):
        self.iter_stop: int = 0
        self.best_epoch: int = 0
        self.best_iter: int = 0

    def _extract_swa_state(
        self,
        model: PyTorchEstimators,
        swa_model: AveragedModel
    ):
        r"""
        Méthode privée qui permet d'extraire les poids du modèle SWA et les 
        renommer pour correspondre aux noms des poids du modèle.

        Parameters
        ----------
        model : PyTorchEstimators
            instance du modèle PyTorch.
        swa_model : AveragedModel
            instance du modèle SWA.
        """
        swa_weights = model._extract_state(swa_model)
        swa_weights = OrderedDict(
            ('.'.join(kk.split('.')[1:]), vv)
            for kk, vv in swa_weights.items() if kk.startswith('module')
        )
        return swa_weights

    def end_eval(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        swa_model: AveragedModel) -> None:
        if (
                res_loss < self.old_res_loss 
                if self.minimize else 
                res_loss > self.old_res_loss
        ):
            if self.verbose:
                print(
                    "\n[SWA] Gain de performance époque/itération : "
                    f"{state.n_epoch_swa+1}/{state.n_iter_swa+1} ;\n"
                    f"Gain de performance {state.loss_name} :\n"
                    f"{self.old_res_loss} -> {res_loss}.\n",
                    flush=True
                )
            self.iter_stop = 0
            self.old_res_loss = res_loss
            self.best_epoch = state.n_epoch_swa
            self.best_iter = state.n_iter_swa
            swa_model.update_parameters(model)
            if self.path_save is None:
                self.state_dict = self._extract_swa_state(model, swa_model)
            else:
                model.save_weights(
                    f"{self.path_save}/{state.id_process}.model",
                    state_dict=self._extract_swa_state(model, swa_model)
                )
        elif (
                self.early_stopping_rounds is not None and 
                (
                    res_loss >= self.old_res_loss 
                    if self.minimize else
                    res_loss <= self.old_res_loss
                )
        ):
            self.iter_stop += 1
            if self.iter_stop >= self.early_stopping_rounds:
                control.training_stop = True

    def end_train(
        self,
        model: PyTorchEstimators,
        data: List[Dict[str, Tensor]],
        state: FitState,
        control: FitControl,
        res_loss: float,
        swa_model: AveragedModel,
        eval_dataset: Optional[DataLoader]
    ) -> None:
        if self.verbose:
            print(
                f"\n[SWA] Meilleure époque/iteration : {self.best_epoch+1}/"
                f"{self.best_iter+1}\n"
                f"Meilleure performance {state.loss_name} : "
                f"{self.old_res_loss}.\n",
                flush=True
            )
        if self.best_iter > 0:
            if self.normalise:
                update_bn(loader=data, swa_model=swa_model);
                if eval_dataset is not None:
                    if self.path_save is None:
                        self.state_dict = self._extract_swa_state(swa_model)
                    else:
                        model.save_weights(
                            f'{self.path_save}/{state.id_processes}.model',
                            state_dict=self._extract_swa_state(swa_model)
                        )
            if self.path_save is None:
                model.load_state_dict(self.state_dict)
            else:
                model.load_weights(
                    f"{self.path_save}/{state.id_process}.model"
                )
                os.remove(f"{self.path_save}/{state.id_process}.model")


@dataclass
class LRSchedulerSWACallback(CallbackInterface, metaclass=SuperclassDocstring):
    r"""
    Classe de gestion du learning rate pour la méthode Stochastic Weight 
    Averaging (SWA).

    Parameters
    ----------
    anneal_epochs : int
        Nombre d'époque pour la phase de recuisson.
    swa_lr : float
        Valeur du learning rate minimal.
    annealing_strategy : str (='cos')
        Stratégie de recuissson :
            - "cos" : stratégie cosinus (par défaut) ;
            - "linear" : stratégie linéaire.
    last_epoch : int (=-1)
        Indice de la dernière époque.

    Notes
    -----
    Cette classe n'a pas pour but d'être utilisée directement, mais est 
    instanciée dans la classe 
    :mod:`~dstk.pytorch._swa.StochasticWeightAveraging`.
    """
    anneal_epochs: int
    swa_lr: float
    anneal_strategy: str = 'cos'
    last_epoch: int = -1

    def begin_train(self,
                    model: PyTorchEstimators,
                    data: List[Dict[str, Tensor]],
                    state: FitState,
                    control: FitControl,
                    res_loss: float,
                    swa_model: AveragedModel) -> None:
        self.scheduler = SWALR(
            optimizer=model.optimizer,
            anneal_epochs=self.anneal_epochs,
            swa_lr=self.swa_lr,
            anneal_strategy=self.anneal_strategy,
            last_epoch=self.last_epoch
        )

    def end_epoch(self,
                  model: PyTorchEstimators,
                  data: List[Dict[str, Tensor]],
                  state: FitState,
                  control: FitControl,
                  res_loss: float,
                  swa_model: AveragedModel) -> None:
        self.scheduler.step()

@dataclass
class StochasticWeightAveraging:
    r"""
    CLasse du processus `Stochastic Weight Averaging (SWA)`_ qui permet de 
    robustifier l'estimateur en moyennant les poids du modèle.

    .. _Stochastic Weight Averaging (SWA): https://arxiv.org/abs/1803.05407

    Parameters
    ----------
    nb_epoch : int
        Nombre d'époques réservé au WSA. Si 0 pas de processus SWA.
    lr : float
        Niveau le plus bas de descente pour le lr scheduler du processus SWA.
    nb_anneal : int (=2)
        Nombre de période fait par le lr scheduler du processus SWA.
    strategy : str (='cos')
        Stratégie du lr scheduler du processus SWA, soit 'cos', ou 'linear'.
    inner_ratio : float (=0.7)
        Ratio d'observation conjoint dans la base d'apprentissage du modèle et 
        du processus SWA. Pour avoir les deux même bases mettre à 1 et pour 
        avoir 2 bases d'apprentissage complétement différentes mettre à 0. 
        Apporter de nouvelles observations pendant le processus SWA permet 
        d'augmenter son efficacité.
    stratified : bool (=False)
        Permet de stratifier les observations communes dans les bases 
        d'apprentissage du train et du processus SWA. Utilisable seulement 
        si le modèle est un classifieur.
    random_state : int (=42)
        random state pour déterminer le mélange des observations pour la 
        construction des deux bases d'apprentissage train et SWA.
    early_stopping_rounds : Optional[int] (=None)
        Nombre d'itérations (epoch) sans amélioration. Si None pas de 
        early stopping, le modèle sauvegardé sera donc le meilleur modèle 
        durant l'ensemble des époques.
    path_save : Optional[str] (=None)
        Chemin (absolu ou relatif) où le modèle par early stopping sera 
        préservé. Si None, les paramètres seront sauvegardés en RAM. 
        Cependant il est déconseillé de le faire car les réseaux profonds 
        peuvent être très gourmands en espace RAM. De même, en cas de 
        plantage, un état du réseau pouvant être satisfaisant est préservé 
        sur le disque dur.
    minimize: bool (=True)
        Indique si une couche de normalisation est présente dans le modèle. 
        Dans ce cas l'évaluation durant l'apprentissage n'est pas possible.
    verbose : bool (=True)
        Indique les informations de gain de performance.

    Notes
    -----
    Le comportement de la classe est de préserver les poids moyennés si le 
    modèle SWA possède de meilleures performances que l'évaluation du modèle 
    de l'évaluation précédente.

    **Aucune de ces méthodes ne doivent être appelées par l'utilisateur**. 
    Elles sont utilisées automatiquement durant le processus d'apprentissage 
    de la méthode :func:`~dstk.pytorch._base.BaseEnvironnement.fit`.

    Warning
    -------
    Si le modèle possède une couche de normalisation alors l'évaluation du 
    modèle au cours de l'apprentissage n'est plus possible. Il sera alors 
    pris à chaque époque l'impact du modèle actuel et une normalisation 
    est appliquée à la fin du processus SWA.
    """
    nb_epoch: int
    lr: float
    nb_anneal: int = 2
    strategy: str = "cos"
    inner_ratio: float = 0.7
    stratified: bool = False
    random_state: int = 42
    early_stopping_rounds: Optional[int] = None
    path_save: Optional[str] = None
    minimize: bool = True
    verbose: bool = True

    def copy(self):
        r"""
        Créer une copie de l'objet.
        """
        return deepcopy(self)

    def init(self,
             model: PyTorchEstimators,
             res_loss: float):
        r"""
        
        """
        self.normalise = any(
            chain.from_iterable(
                map(
                    lambda y: map(
                        lambda x: f'{x}' in f'{y[1]}',
                        BATCH_NORM_LAYERS
                    ),
                    model.named_children()
                )
            )
        )
        if self.normalise:
            warn(
                "Une couche de Batch Normalization a été détectée. Il "
                "ne sera pas possible d'évaluer les performances du "
                "processus SWA durant les époques. Donc toutes les "
                "itérations seront prises en compte dans le modèle "
                "final."
            )
        callbacks = [
            SaveStateSWACallback(
                old_res_loss=res_loss,
                path_save=self.path_save,
                normalise=self.normalise,
                minimize=self.minimize,
                early_stopping_rounds=self.early_stopping_rounds,
                verbose=self.verbose
            ),
            LRSchedulerSWACallback(
                anneal_epochs=self.nb_epoch,
                swa_lr=self.lr,
                anneal_strategy=self.strategy,
                last_epoch=-1
            )
        ]
        if self.verbose:
            callbacks.append(ProgressBarCallback(_swa_process=True))
        self.callbacks = CallbackHandler(callbacks)
        self._estimator_type = model._estimator_type

    def _divergence(self, loss: float):
        r"""
        Méthode interne d'alerte de divergence du modèle pendant sa phase 
        d'apprentissage.
        """
        if np.isnan(loss) or np.isinf(loss):
            raise DivergenceError(
                """Verger ! L'algorithme a dit "verger" !"""
            )

    def _eval_step(self,
                   swa_model: AveragedModel,
                   data: List[Dict[str, Any]],
                   field_target: str,
                   loss_fn: Callable,
                   kargs_loss: Dict[str, Any]):
        r"""
        Méthode privée d'une étape d'évaluation.

        Parameters
        ----------
        swa_model : AveragedModel
            Modèle PyTorch d'AveragedModel.
        data : List[Dict[str, Any]]
            Liste de dictionnaires sur les données d'un mini-batch.
        field_target : str
            Nom du champs de la target.
        loss_fn : Callable
            Fonction coût PyTorch.
        kargs_loss : Dict[str, Any]
            Dictionnaire d'arguments à utiliser dans la fonction de coût 
            loss_fun.

        Returns
        -------
        loss : float
            Valeur de la fonction coût au travers du mini-batch.
        """
        swa_model.eval()
        with no_grad():
            output = swa_model.forward(**data)
            target = check_tensor(data[field_target])
            loss = loss_fn(output, target, **kargs_loss)
        loss = loss.detach()
        self._divergence(loss)
        return loss

    def _iter_eval(self,
                   model: PyTorchEstimators,
                   swa_model: AveragedModel,
                   eval_dataset: DataLoader,
                   field_target: str,
                   loss_fn: Callable,
                   kargs_loss: Optional[Dict[str, Any]],
                   state: FitState,
                   control: FitControl,
                   list_callbacks: List[CallbackInterface]):
        r"""
        Méthode interne exécutant la phase d'évaluation durant l'apprentissage 
        du processus SWA.Cette étape peut être appelée à la fin d'une époque 
        ou durant l'exécution de l'époque (dépend de l'argument iter_eval de 
        la méthode :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).

        Parameters
        ----------
        model : PyTorchEstimators
            Modèle d'origine.
        swa_model : AveragedModel
            Modèle PyTorch d'AveragedModel.
        eval_dataset : DataLoader
            Dataset d'évaluation au format DataLoeader généré dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`.
        field_target : str
            Nom du champs target.
        loss_fn : Callable
            Fonction coût à minimiser (utiliser les fonctions présentes dans 
            torch.nn.functional). Si la fonction est une fonction custom, elle 
            doit avoir le prototypage suivant :
                - loss_fn(y_estim, y_true, reduction='sum', **kargs)
            avec une implémentation de réduction 'sum' pour les phases 
            d'évaluation et 'mean' pour les phases d'apprentissage.
        kargs_loss : Optional[Dict[str, Any]]
            Dictionnaire d'arguments à utiliser dans la fonction de coût 
            loss_fun.
        state : FitState
            Etat de l'apprentissage (instancié dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        control : FitControl
            Contrôle de l'apprentissage (instancié dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        list_callbacks : List[CallbackInterface]
            Liste des callback permettant de contôler l'apprentissage durant 
            la phase d'évaluation.
        """
        swa_model_int = deepcopy(swa_model)
        swa_model_int.update_parameters(model)
        list_callbacks.begin_eval(
            model=model,
            data=eval_dataset,
            state=state,
            control=control,
            res_loss=None,
            swa_model=swa_model
        )
        loss_eval = 0
        for jj, eval_batch in enumerate(eval_dataset):
            list_callbacks.begin_eval_step(
                model=model,
                data=eval_batch,
                state=state,
                control=control,
                res_loss=loss_eval/(jj+1),
                swa_model=swa_model
            )
            try:
                loss_eval += self._eval_step(
                    swa_model=swa_model_int,
                    data=eval_batch,
                    field_target=field_target,
                    loss_fn=loss_fn,
                    kargs_loss=kargs_loss
                )
            except DivergenceError as error:
                control.diverge = True
                warn(error.msg)
                break
        list_callbacks.end_eval(
            model=model,
            data=eval_dataset,
            state=state,
            control=control,
            res_loss=loss_eval/(jj+1),
            swa_model=swa_model
        )

    def fit(self,
            model: PyTorchEstimators,
            eval_dataset: Optional[DataLoader],
            loss_fn: Callable,
            kargs_loss: Optional[Dict[str, Any]],
            field_target: str,
            iter_eval: Optional[int],
            collate_fn: Callable,
            batch_size: int,
            shuffle: bool,
            drop_last: bool,
            num_workers: int,
            state: FitState,
            control: FitControl):
        r"""
        Processus d'apprentissage SWA.
        
        Parameters
        ----------
        model : PyTorchEstimators
            Modèle d'origine.
        eval_dataset : Optional[DataLoader]]
            DataLoader PyTorch contenant les observations et les targets 
            d'évaluation. Si eval_set est None alors pas de early stopping 
            et évaluation des performances sur les données d'entrainement 
            (déconseillé car très fort risque d'overfitting).
        loss_fn : Optional[Callable] (=None)
            Fonction coût à minimiser (utiliser les fonctions présentes dans 
            torch.nn.functional). Si la fonction est une fonction custom, elle 
            doit avoir le prototypage suivant :
                - loss_fn(y_estim, y_true, reduction='sum', **kargs)
            avec une implémentation de réduction 'sum' pour les phases 
            d'évaluation et 'mean' pour les phases d'apprentissage.
        kargs_loss : Dict[str, Any] (=dict())
            Dictionnaire d'arguments à utiliser dans la fonction de coût 
            loss_fun. Si il n'y a pas d'arguments, mettre un dictionnaire 
            vide.
        field_target : str
            Nom du champ contenu dans X de la target. Par défaut "target". 
            Attention l'implémentation de la fonction collate_fn peut faire 
            changer le nom de la target.
        iter_eval : Optional[int]
            Nombre d'itération entre deux évaluations. Actif seulement si 
            eval_set est non None. Si iter_eval est à None (défini par 
            défaut) les évaluations du modèle seront à la fin des époques. 
            Si un entier positif est renseigné les évaluations auront lieu 
            tous les iter_eval itérations d'apprentissage.
        collate_fn : Optional[Callable]
            Fonction de sortie de DataLoader permettant la concaténation en 
            batch des différentes observations du mini-batch.
        batch_size : int
            taille du mini-batch. Pris en compte seulement si dataloader = 
            True.
        shuffle : bool 
            Permet de mélanger à chaque époque les observations. Pris en 
            compte seulement si dataloader = True.
        drop_last : bool
            Drop les dernières observations si le rapport entre le nombre 
            d'observations et la taille du mini-batch est non entier. Pris en 
            compte seulement si dataloader = True.
        num_workers : int
            Nombre de process du DataLoader. Par défaut 0 correspondant à du 
            mono-thread. Pris en compte seulement si dataloader = True.
                - Attention :
                    1 -> mono-threadé mais isolé de l'environnement 
                    (Picklelisé). Si le DataLoader est utilisé dans un 
                    environnement Sciki-Learn multi-processing (par exemple 
                    pour un RandomizedSearchCV) il est très important de 
                    mettre cet argument à 0 pour que la DataLoader tourne 
                    dans l'environnement principal de Python. Dans le cas 
                    contraire le processus plantera.
        state : FitState
            Etat de l'apprentissage (instancié dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        control : FitControl
            Contrôle de l'apprentissage (instancié dans la méthode 
            :func:`~dstk.pytorch._base.BaseEnvironnement.fit`).
        """
        if self.verbose:
            print(
                "Début de la procédure de Stochastic Weight Averagin (SWA) "
                "avec :\n"
                f"    - {self.nb_epoch} époques ;\n"
                f'    - {self.nb_anneal} nombres de "recuissons" ;\n'
                f"    - {self.strategy} stratégies ;\n"
                f"    - un ratio d'intersection de {self.inner_ratio}."
            )
        swa_model = AveragedModel(model)
        swa_model.update_parameters(model)

        if not isinstance(self.X_swa, DataLoader):
            dataset = DataLoader(
                self.X_swa,
                collate_fn=collate_fn,
                shuffle=shuffle,
                drop_last=drop_last,
                batch_size=batch_size,
                num_workers=num_workers
            )
        else:
            dataset = self.X_swa

        state.nb_epoch_swa = self.nb_epoch
        self.callbacks.begin_train(
            model=model,
            data=dataset,
            state=state,
            control=control,
            res_loss=None,
            swa_model=swa_model
        )
        for n_epoch in range(self.nb_epoch):
            state.n_epoch_swa = n_epoch
            loss_epoch = 0
            self.callbacks.begin_epoch(
                model=model,
                data=dataset,
                state=state,
                control=control,
                res_loss=loss_epoch,
                swa_model=swa_model
            )
            for ii, batch in enumerate(dataset):
                state.n_iter_swa = ii
                if (
                    eval_dataset and
                    model._mod(n_epoch*state.epoch_max_steps+ii, iter_eval) == 0
                ):
                    self._iter_eval(
                        model=model,
                        swa_model=swa_model,
                        eval_dataset=eval_dataset,
                        field_target=field_target,
                        loss_fn=loss_fn,
                        kargs_loss=kargs_loss,
                        state=state,
                        control=control,
                        list_callbacks=self.callbacks
                    )
                    if control.diverge or control.training_stop:
                        break
                self.callbacks.begin_step(
                    model=model,
                    data=batch,
                    state=state,
                    control=control,
                    res_loss=None,
                    swa_model=swa_model
                )
                try:
                    loss_step = model._train_step(
                        data=batch,
                        field_target=field_target,
                        loss_fn=loss_fn,
                        kargs_loss=kargs_loss
                    )
                    loss_epoch += loss_step
                    self.callbacks.end_step(
                        model=model,
                        data=batch,
                        state=state,
                        control=control,
                        res_loss=loss_step,
                        swa_model=swa_model
                    )
                except DivergenceError as error:
                    control.diverge = True
                    warn(error.msg)
                    break
            loss_epoch /= (ii+1)
            self.callbacks.end_epoch(
                model=model,
                data=dataset,
                state=state,
                control=control,
                res_loss=loss_epoch,
                swa_model=swa_model
            )
            if eval_dataset and iter_eval is None:
                self._iter_eval(
                    model=model,
                    swa_model=swa_model,
                    eval_dataset=eval_dataset,
                    field_target=field_target,
                    loss_fn=loss_fn,
                    kargs_loss=kargs_loss,
                    state=state,
                    control=control,
                    list_callbacks=self.callbacks
                )
            if control.diverge or control.training_stop:
                break
        self.callbacks.end_train(
            model=model,
            data=dataset,
            state=state,
            control=control,
            res_loss=None,
            swa_model=swa_model,
            eval_dataset=eval_dataset
        )

def cut_dataset(
    model: PyTorchEstimators,
    data: Union[Dataset, DataLoader, List[Dict[str, Any]]],
    *,
    ratio: float=0.5,
    stratified: bool=False,
    target_field: str='target',
    random_state: int=42
):
    r"""
    Fonction qui découpe un dataset en deux datasets en fonction d'un taux de 
    recouvrement des observations.

    Parameters
    ----------
    model: PyTorchEstimators
        Modèle d'apprentissage PyTorch.
    data : Union[Dataset, DataLoader, List[Dict[str, Any]]]
        Dataset ou list de dictionnaire des observations.
    ratio : float (=0.5)
        Taux de recouvrement des observations entre les deux datasets.
    stratified : bool (=False)
        Permet de stratifier la répartition en fonction des classes de la 
        target (ne fonctionne que si il s'agit d'un classifieur).
    target_field : str (='target')
        Précise le champs target si il y a une stratification (sinon non 
        utilisé).
    random_state : int (=42)
        Random state de la répartition des observations.

    Returns
    -------
    DatasetTrain : Union[Datast, Dataset, List[Dict[str, Tensor]]]
        Dataset utilisé pour la phase d'entrainement
    """
    if ratio < 0 or ratio > 1:
        raise AttributeError(
            "swa_inner_ratio doit être compris entre 0 et 1 "
            f"(inclus). Ici il vaut {ratio} !"
        )
    if stratified and model._estimator_type != 'classifier':
        raise AttributeError(
            "Pour l'approche stratifiée de l'intersection jointe il faut "
            "que le modèle soit un classifier, or il s'agit d'une "
            f"{model._estimator_type} !"
        )
    if isinstance(data, DataLoader):
        is_dataloader = True
        dl_args = {
            kk: vv for kk, vv in data.__dict__.items()
            if not kk.startswith('_')
            and not kk in ('dataset', 'batch_sampler')
        }
        data = data.dataset
    else:
        is_dataloader = False
    size = len(data)
    idx_ = np.arange(size)
    y = None
    if ratio < 1 and ratio > 0:
        if stratified:
            split = StratifiedShuffleSplit(
                n_splits=1,
                train_size=int(ratio*size),
                random_state=random_state
            )
            y = list(map(lambda x: x[target_field], data))
        else:
            split = ShuffleSplit(
                n_splits=1,
                train_size=int(ratio*size),
                random_state=random_state
            )
        idx_over, idx_rest = next(split.split(idx_, y))
    elif ratio == 1:
        idx_over, idx_rest = idx_, np.array([])
    elif ratio == 0:
        idx_over, idx_rest = np.array([]), idx_
    idx_train = np.r_[
        idx_rest[:int(idx_rest.shape[0]//2)],
        idx_over
    ]
    idx_swa = np.r_[
        idx_rest[int(idx_rest.shape[0]//2):],
        idx_over
    ]
    if is_dataloader:
        return (
            DataLoader(dataset=Subset(data, idx_train), **dl_args),
            DataLoader(dataset=Subset(data, idx_swa), **dl_args)
        )
    elif isinstance(data, Dataset):
        return Subset(data, idx_train), Subset(data, idx_swa)
    else:
        return (
            [data[ii] for ii in idx_train],
            [data[ii] for ii in idx_swa]
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Permet de vérifier si on accède a l'__init__.py via le setup.
try:
    __SETUP__
except NameError:
    __SETUP__ = False

# teste si pytorch est installé si ce n'est pas via le setup qu'on passe par 
# le __init__.py !
if __SETUP__:
    pass
else :
    try:
        import torch
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            "Le module PyTorch n'est pas installé : l'instruction import torch "
            "renvoit une erreur ModuleNotFoundError. Visitez la page du projet "
            "PyTorch pour savoir comment l'installer."
        )

    from dstk.pytorch._utils import (check_tensor, check_tensor_dict, detach,
                                     collate_zeros_padding, make_mask)
    from dstk.pytorch._callback import (CallbackInterface, CallbackHandler,
                                        FitState, FitControl, PrintCallback,
                                        ProgressBarCallback,
                                        LRSchedulerCallback,
                                        EarlyStoppingCallback, Supervision,
                                        DEFAULT_CALLBACKS)
    from dstk.pytorch._swa import StochasticWeightAveraging, cut_dataset
    from dstk.pytorch._random_search import RandomizedSearchOnline
    from dstk.pytorch._classifier import BaseClassifier, BaseClassifierOnline
    from dstk.pytorch._regressor import BaseRegressor, BaseRegressorOnline
    
    __all__ = (
        'BaseClassifier',
        'BaseClassifierOnline',
        'BaseRegressor',
        'BaseRegressorOnline',
        'CallbackInterface',
        'CallbackHandler',
        'FitState',
        'FitControl',
        'PrintCallback',
        'ProgressBarCallback',
        'EarlyStoppingCallback',
        'LRSchedulerCallback',
        'Supervision',
        'DEFAULT_CALLBACKS',
        'StochasticWeightAveraging',
        'cut_dataset',
        'check_tensor',
        'check_tensor_dict',
        'detach',
        'RandomizedSearchOnline',
        'collate_zeros_padding',
        'make_mask'
    )
